package main

import (
	"fmt"
	"sync"
)

type New func() interface{}

type Pool struct {
	mutex   sync.Mutex
	objects []interface{} // 切片
	new     New
}

func NewPool(new New) *Pool {
	return &Pool{
		objects: make([]interface{}, 0),
		new:     new,
	}
}

func NewSizePool(new New, size int) *Pool {
	// 初始化池的数量
	objects := make([]interface{}, size)
	for i := 0; i < size; i++ {
		objects[i] = new()
	}

	return &Pool{
		objects: objects,
		new:     new,
	}
}

func (p *Pool) Get() interface{} {
	p.mutex.Lock()
	defer p.mutex.Unlock()

	var result interface{}

	if len(p.objects) > 0 {
		// 队列方式获取
		result = p.objects[0]
		p.objects = p.objects[1:]
		// 堆栈方式获取
		//l := len(p.objects)
		//result = p.objects[l-1]
		//p.objects = p.objects[:l-1]

		return result
	} else {
		// 如果没有就新建
		return p.new()
	}
}

func (p *Pool) Put(obj interface{}) {
	p.mutex.Lock()
	defer p.mutex.Unlock()

	p.objects = append(p.objects, obj)
}

func main() {
	p := NewPool(func() interface{} {
		fmt.Println("call new")
		return 1
	})

	// 从池里获取一个
	x := p.Get()
	fmt.Printf("%T %#v\n", x, x)

	// 放回池
	p.Put(x)

	// 再次从池里获取，不调用新建
	x = p.Get()
	fmt.Printf("%T %#v\n", x, x)

	// 在调用会创建，池里已经没有了，会重新创建
	x = p.Get()
	fmt.Printf("%T %#v\n", x, x)
	fmt.Println()

	p = NewSizePool(func() interface{} {
		fmt.Println("call new 2")
		return 2
	}, 2)
	for i := 0; i < 3; i++ {
		x = p.Get()
		fmt.Printf("%T %#v\n", x, x)
	}

	// Output:
	//call new
	//int 1
	//int 1
	//call new
	//int 1
	//
	//call new 2
	//call new 2
	//int 2
	//int 2
	//call new 2
	//int 2
}
